/*
    Copyright (c) 2010, 2011, 2012 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>
            Michael Troß, aizo GmbH <michael.tross@aizo.com>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
    #include "autoconfig.h"
#endif

#include <boost/pointer_cast.hpp>
#include <errno.h>
#include <string.h>
#include <exception>
#include <iomanip>

#include "common.h"
#include "log.h"
#include "util.h"
#include "exceptions.h"
#include "config_manager.h"
#include "hwdesc.h"
#include "hwstats.h"
#include "storage.h"
#include "device-config.h"
#include "database.h"
#include "ds-assistant.h"
#include "database.h"
#include "util.h"
#include "traffic-decode.h"
#include "sqlite3_wrapper.h"
#include "gettext_helper.h"
#include "ds_util.h"
#include "comm-channel.h"


namespace dSA
{

dS_Assistant* dS_Assistant::m_instance = NULL;

dS_Assistant* dS_Assistant::createInstance()
{
    if (m_instance == NULL)
    {
        m_instance = new dS_Assistant();
        if (m_instance == NULL)
        {
            throw _Exception("could not create dS Assistant class");
        }
    }

    return m_instance;
}

dS_Assistant* dS_Assistant::getInstance()
{
    if (m_instance == NULL)
    {
        throw _Exception("dS Assistant not initialized");
    }

    return m_instance;
}
dS_Assistant::dS_Assistant() : m_shutdown_flag(false),
                               m_running(false),
                               m_task_thread(0),
                               m_long_task_thread(0)
{
    int ret;

    ConfigManager *cfg = ConfigManager::getInstance();
    std::string logdir(cfg->getOption(CFG_LOG_DIR));
    m_logfile_name = logdir +  + "/dsa.log";
    m_logfile = fopen(m_logfile_name.c_str(), "a");
    if (NULL == m_logfile)
    {
        FILE_LOG(logERROR) << "logs will not be written, failed to open " <<
                              "logfile \'" << m_logfile_name << "\': " <<
                              strerror(errno);
    } else {
        Output2DSUID::SetLogdirectory(logdir);
        Output2FILE::SetStream(m_logfile);
    }

    DSUIDLog::ReportingLevel() =
        DSUIDLog::FromString((cfg->getOption(CFG_LOG_LEVEL)).c_str());

    FILELog::ReportingLevel() =
        FILELog::FromString((cfg->getOption(CFG_LOG_LEVEL)).c_str());
    FILE_LOG(logINFO) << "dS Assistant startup";

    SetNullDsuid(m_own_dsuid);

    ret = pthread_mutex_init(&m_task_list_mutex, NULL);
    if (ret != 0)
    {
        throw _Exception("failed to initialize task list mutex: " +
                         std::string(strerror(errno)));
    }

    ret = pthread_mutex_init(&m_long_task_list_mutex, NULL);
    if (ret != 0)
    {
        pthread_mutex_destroy(&m_task_list_mutex);
        throw _Exception("failed to initialize long task list mutex: " +
                         std::string(strerror(errno)));
    }

    ret = pthread_mutex_init(&m_suspended_tasks_mutex, NULL);
    if (ret != 0)
    {
        pthread_mutex_destroy(&m_task_list_mutex);
        pthread_mutex_destroy(&m_long_task_list_mutex);
        throw _Exception("failed to initialize suspended tasks mutex: " +
                         std::string(strerror(errno)));
    }


    ret = pthread_cond_init(&m_wakeup_task_cond, NULL);
    if (ret != 0)
    {
        pthread_mutex_destroy(&m_task_list_mutex);
        pthread_mutex_destroy(&m_long_task_list_mutex);
        pthread_mutex_destroy(&m_suspended_tasks_mutex);
        throw _Exception("failed to initialize task thread wakeup condition: " +
                         std::string(strerror(errno)));
    }

    ret = pthread_cond_init(&m_wakeup_long_task_cond, NULL);
    if (ret != 0)
    {
        pthread_mutex_destroy(&m_task_list_mutex);
        pthread_mutex_destroy(&m_long_task_list_mutex);
        pthread_cond_destroy(&m_wakeup_task_cond);
        pthread_mutex_destroy(&m_suspended_tasks_mutex);
        throw _Exception("failed to initialize task thread wakeup condition: " +
                         std::string(strerror(errno)));
    }
}

void dS_Assistant::reopenLog()
{
    Output2FILE::SetStream(NULL);
    if (NULL != m_logfile)
    {
        fclose(m_logfile);
    }
    m_logfile = fopen(m_logfile_name.c_str(), "a");
    if (NULL == m_logfile)
    {
        throw _Exception("failed to re-open logfile \'" + m_logfile_name + "\': " +
                         std::string(strerror(errno)));
    }
    Output2FILE::SetStream(m_logfile);
}

void dS_Assistant::loadDatabaseDefaults()
{
    int apartment_id;
    std::string apartment_id_string;

    // system parameter: apartment id for automatic meter configuration
    try
    {
        ConfigManager::getInstance()->getParameter(
                "user.digitalstrom.apartmentid", apartment_id);
    }
    catch (Exception& ex )
    {
        try
        {
            // migrate old values from database if missing in config file
            apartment_id_string = Database::getInstance()->getConfig("ApartmentId");
            ConfigManager::getInstance()->setParameter(
                    "user.digitalstrom.apartmentid",
                    util::str2int(apartment_id_string));
        }
        catch (Exception& ex )
        {
            // option is neither in config file nor in database
            std::string systemid = util::mac2str();
            char c = systemid.length() ? systemid[systemid.length() - 1] : '0';
            apartment_id = (strtol(&c, NULL, 16) & 7) * 8;

            FILE_LOG(logINFO) << "Set apartment id from system mac address "
                    << systemid << " to " << util::str(apartment_id);
            try
            {
                ConfigManager::getInstance()->setParameter(
                        "user.digitalstrom.apartmentid", apartment_id);
            }
            catch (Exception& ex)
            {
                FILE_LOG(logERROR)
                        << "Failed to store apartment id in configuration: "
                        << ex.getMessage();
            }
        }
    }

    int value;
    bool bvalue;

    // meter configuration: default behavior for checking the device presence
    try
    {
        ConfigManager::getInstance()->getParameter(
                "digitalstrom.meter.pingmode", value);
    }
    catch (Exception& ex)
    {
        try
        {
            value = util::str2int(Database::getInstance()->getConfig(
                    "MeterPingMode"));
        }
        catch (Exception&) {
            value = 3;
        }
        try
        {
            ConfigManager::getInstance()->setParameter(
                    "digitalstrom.meter.pingmode", value);
        }
        catch (Exception& ex)
        {
            FILE_LOG(logERROR)
                << "Failed to store meter.pingmode in configuration: "
                << ex.getMessage();
        }
    }

    // meter configuration: default warning temperature level
    try
    {
        ConfigManager::getInstance()->getParameter(
                "digitalstrom.meter.twarning", value);
    }
    catch (Exception& ex )
    {
        try
        {
            value = util::str2int(Database::getInstance()->getConfig(
                    "MeterTemperatureWarning"));
        }
        catch (Exception&)
        {
            value = 90;
        }
        try
        {
            ConfigManager::getInstance()->setParameter(
                    "digitalstrom.meter.twarning", value);
        }
        catch (Exception& ex)
        {
            FILE_LOG(logERROR)
                << "Failed to store meter.twarning in configuration: "
                << ex.getMessage();
        }
    }

    // meter configuration: default error temperature level
    try
    {
        ConfigManager::getInstance()->getParameter(
                "digitalstrom.meter.tfailure", value);
    }
    catch (Exception& ex ) {
        try
        {
            value = util::str2int(Database::getInstance()->getConfig(
                    "MeterTemperatureError"));
        }
        catch (Exception&) {
            value = 150;
        }
        try
        {
            ConfigManager::getInstance()->setParameter(
                    "digitalstrom.meter.tfailure", value);
        }
        catch (Exception& ex)
        {
            FILE_LOG(logERROR)
                << "Failed to store meter.tfailure in configuration: "
                << ex.getMessage();
        }
    }

    // meter configuration: front led configuration enabled/disabled in idle state
    try
    {
        ConfigManager::getInstance()->getParameter(
                "user.digitalstrom.meter.led", bvalue);
    }
    catch (Exception& ex ) {
        bvalue = true;
        try
        {
            ConfigManager::getInstance()->setParameter(
                    "user.digitalstrom.meter.led", bvalue);
        }
        catch (Exception& ex)
        {
            FILE_LOG(logERROR)
                << "Failed to store meter.led in configuration: "
                << ex.getMessage();
        }
    }

    // System parameter: programming mode enabled/disabled
    try
    {
        ConfigManager::getInstance()->getParameter(
                "user.digitalstrom.programmingmode", bvalue);
    }
    catch (Exception& ex )
    {
        bvalue = true;
        try
        {
            ConfigManager::getInstance()->setParameter(
                    "user.digitalstrom.programmingmode", bvalue);
        }
        catch (Exception&) {
            FILE_LOG(logERROR) << "Failed to store programmingmode "
                    "parameter in configuration: " << ex.getMessage();
        }
    }

    // System parameter: room dimming mode enabled/disabled
    try
    {
        ConfigManager::getInstance()->getParameter(
                "user.digitalstrom.roomdimmingmode", bvalue);
    }
    catch (Exception& ex )
    {
        bvalue = true;
        try
        {
            ConfigManager::getInstance()->setParameter(
                    "user.digitalstrom.roomdimmingmode", bvalue);
        }
        catch (Exception&) {
            FILE_LOG(logERROR) << "Failed to store roomdimmingmode "
                    "parameter in configuration: " << ex.getMessage();
        }
    }

}

void dS_Assistant::connect()
{
    boost::mutex::scoped_lock lock(m_dsa_mutex);

    if (m_dsm_api_wrapper != NULL)
    {
        throw _Exception("connection already exists");
    }

    ConfigManager *cfg = ConfigManager::getInstance();
    m_dsm_api_wrapper = boost::shared_ptr<dSMApiWrapper>(
                    new dSMApiWrapper(cfg->getOption(CFG_DS485_DAEMON_URL)));

    try
    {
        m_dsm_api_wrapper->getOwnDSUID(&m_own_dsuid);
    }
    catch (Exception &ex)
    {
        m_dsm_api_wrapper.reset();
        throw ex;
    }

    m_bus_state = m_dsm_api_wrapper->getBusState();

    m_dsm_api_wrapper->addListener(this);
}

bus_state_t dS_Assistant::getBusState() const
{
    boost::mutex::scoped_lock lock(m_dsa_mutex);
    return m_bus_state;
}

boost::shared_ptr<dSMApiWrapper> dS_Assistant::dS485()
{
    if (m_dsm_api_wrapper == NULL)
    {
        throw _Exception("no active connection to dS485 daemon available");
    }
    return m_dsm_api_wrapper;
}

void dS_Assistant::addTask(boost::shared_ptr<dSTask> task)
{
    if (task == NULL)
    {
        throw _Exception("Will not add invalid NULL task");
    }

    FILE_LOG(logDEBUG2) << "adding task " << task->getName();

    lockTaskList();
    m_task_list.push_front(task);

    pthread_cond_signal(&m_wakeup_task_cond);
    unlockTaskList();
}

long long int dS_Assistant::addLongTask(boost::shared_ptr<dSLongTask> task)
{
    FILE_LOG(logDEBUG2) << "adding long task " <<
            task->getName() << "/" << task->getTaskID();

    lockLongTaskList();

    long long int task_id = INVALID_TASK_ID;

    try
    {
        // persistent tasks always have the lower piority and will never
        // cancel currently running tasks
        if ((m_current_task_info != NULL) && (!task->isPersistent()))
        {
            if ((m_current_task_info->getPriority() == Normal) &&
                (task->getPriority() == High))
            {
                if (m_current_task_info->isPausable())
                {
                    FILE_LOG(logDEBUG2) << "pausing current task "
                        << m_current_task_info->getName() << "/" << task->getTaskID();
                    m_current_task_info->pause();
                }
                else
                {
                    FILE_LOG(logDEBUG2)
                        << "current task is unpausable and will be cancelled: " <<
                        m_current_task_info->getName() << "/" << task->getTaskID();
                    m_current_task_info->invalidate();
                }
            }
            else if ((m_current_task_info->getPriority() == High) &&
                     (task->getPriority() == High))
            {
                FILE_LOG(logDEBUG2) << "replacing current high prio task: "
                    << m_current_task_info->getName() << "/" << task->getTaskID();
                m_current_task_info->invalidate();
            }
        }

        task->setTaskStatus(Queued);
        task_id = task->getTaskID();

        if (task_id == INVALID_TASK_ID)
        {
            task_id = Database::getInstance()->saveTask(task);
            task->setTaskID(task_id);
        }
        else
        {
            if (task->getTaskType() !=
                Database::getInstance()->getTaskType(task_id))
            {
                throw _Exception("task type mismatch");
            }
        }

        // prevent programming mistakes
        if ((task->getPriority() == High) && task->isPersistent())
        {
            throw _Exception("invalid persistent task configuration");
        }

        if (task->getPriority() == High)
        {
            // we agreed to allow only one high prio task for now...
            while (!m_high_prio_task_list.empty())
            {
                boost::shared_ptr<dSLongTask> t = m_high_prio_task_list.front();
                t->invalidate();
                m_high_prio_task_list.pop_front();
            }
            FILE_LOG(logDEBUG2) << "adding high priority task: "
                << task->getName() << "/" << task->getTaskID();
            m_high_prio_task_list.push_front(task);
        }
        else
        {
            if (task->isPersistent()) 
            {
                // only one task of the same type is allowed
                bool found = false;
                std::list<boost::shared_ptr<dSLongTask> >::iterator it =
                    m_long_task_list.begin();

                while (it != m_long_task_list.end())
                {
                    boost::shared_ptr<dSLongTask> t = *it;
                    if (task->getTaskType() == t->getTaskType())
                    {
                        found = true;
                        it = m_long_task_list.erase(it);
                        break;
                    }
                    it++;
                }
                if (!found)
                {
                    it = m_suspended_task_list.begin();

                    while (it != m_suspended_task_list.end())
                    {
                        boost::shared_ptr<dSLongTask> t = *it;
                        if (task->getTaskType() == t->getTaskType())
                        {
                            it = m_suspended_task_list.erase(it);
                            break;
                        }
                        it++;
                    }
                }

                // check if task is currently in suspended mode or not
                if (isTaskSuspended(task->getTaskType())) {
                    FILE_LOG(logDEBUG2) <<
                        "adding suspended persistent task: " <<
                        task->getName() << "/" << task->getTaskID();
                    m_suspended_task_list.push_front(task);
                } else {
                    FILE_LOG(logDEBUG2) << "adding persistent task: " <<
                        task->getName() << "/" << task->getTaskID();
                    m_long_task_list.push_front(task);
                }
            }
            else
            {
                FILE_LOG(logDEBUG2) << "adding normal priority task: " <<
                    task->getName() << "/" << task->getTaskID();
                m_long_task_list.push_front(task);
            }

        }

        pthread_cond_signal(&m_wakeup_long_task_cond);
    }
    catch (Exception &ex)
    {
        unlockLongTaskList();
        throw ex;
    }

    unlockLongTaskList();
    return task_id;
}

bool dS_Assistant::isTaskSuspended(task_type type) const
{
    lockSuspendedTaskInfo();
    bool isSuspended = false;
    try
    {
        if (!m_suspended_tasks.empty())
        {
            std::list<task_type>::const_iterator it = m_suspended_tasks.begin();
            while (it != m_suspended_tasks.end())
            {
                if (*it == type)
                {
                    isSuspended = true;
                    break;
                }
                it++;
            }
        }
    }
    catch (...)
    {
        FILE_LOG(logERROR) << "exception caught when checking suspended task"; 
    }
    unlockSuspendedTaskInfo();
    return isSuspended;
}

void dS_Assistant::setSuspendedTask(task_type type)
{
    if (!isTaskSuspended(type))
    {
        lockSuspendedTaskInfo();
        m_suspended_tasks.push_back(type);
        unlockSuspendedTaskInfo();
    }
}

void dS_Assistant::unsetSuspendedTask(task_type type)
{
    lockSuspendedTaskInfo();
    try
    {
        if (!m_suspended_tasks.empty())
        {
            std::list<task_type>::iterator it = m_suspended_tasks.begin();
            while (it != m_suspended_tasks.end())
            {
                if (*it == type)
                {
                    m_suspended_tasks.erase(it);
                    break;
                }
                it++;
            }
        }
    }
    catch (...)
    {
        FILE_LOG(logERROR) << "exception caught when checking suspended task";
    }
    unlockSuspendedTaskInfo();
}


void dS_Assistant::scheduleTask(boost::shared_ptr<dSTask> task,
                                unsigned int seconds)
{
    lockTaskList();

    FILE_LOG(logDEBUG2) << "scheduleTask: " << task->getName()
        << ", delay = " << seconds;

    try
    {
        boost::shared_ptr<TaskScheduler> ts(new TaskScheduler(task, seconds));
        Timer::getInstance()->addSubscriber(ts);
    }
    catch (Exception &ex)
    {
        unlockTaskList();
        throw ex;
    }
    unlockTaskList();
}

void dS_Assistant::scheduleLongTask(boost::shared_ptr<dSLongTask> task,
                           unsigned int seconds)
{
    boost::shared_ptr<LongTaskScheduler> ts;

    // do not schedule already suspended tasks
    if (isTaskSuspended(task->getTaskType()))
    {
        return;
    }

    lockLongTaskList();

    FILE_LOG(logDEBUG2) << "scheduleLongTask: "
        << task->getName() << "/" << task->getTaskID()
        << ", delay = " << seconds;

    try
    {
        // check and clean up stale tasks
        std::list<boost::weak_ptr<LongTaskScheduler> >::iterator wi =
            m_scheduled_task_list.begin();
        while (wi != m_scheduled_task_list.end())
        {
            boost::weak_ptr<LongTaskScheduler> tw = *wi;
            if (!tw.lock())
            {
                FILE_LOG(logDEBUG2)
                                << "scheduleLongTask: cleaned up stale task";
                wi = m_scheduled_task_list.erase(wi);
                continue;
            }
            wi++;
        }

        if (task->getTaskStatus() == Sleeping)
        {
            unlockLongTaskList();
            return;
        }

        task->setTaskStatus(Queued);
        task->setQueuedTime(seconds);
        ts = boost::shared_ptr<LongTaskScheduler> (new LongTaskScheduler(task,
            seconds));
        boost::weak_ptr<LongTaskScheduler> tsw(ts);
        m_scheduled_task_list.push_back(tsw);
    }
    catch (Exception &ex)
    {
        unlockLongTaskList();
        throw ex;
    }
    unlockLongTaskList();
    Timer::getInstance()->addSubscriber(ts);
}

///****************************************************************************

void dS_Assistant::busStateCallback(bus_state_t state)
{
    m_bus_state = state;
    FILE_LOG(logINFO) << "dS485 Bus State: " << state;

    if (state == DS485_STATE_ISOLATED || state == DS485_STATE_DISCONNECTED)
    {
        try
        {
            /// \todo do not access database in callbacks
            Database::getInstance()->updateDSMPresence(false);
        }
        catch (SQLite3Exception &sx)
        {
            FILE_LOG(logERROR) << "Could not update dSM presence: " <<
                    sx.getMessage();
            if (SQLite3::isFatal(sx.getErrorCode()))
            {
                FILE_LOG(logERROR) << "Unrecoverable SQLite error, terminating dSA";
                g_shutdown_flag = SHUTDOWN_SQLITE;
            }
        }
    }
}


void dS_Assistant::busChangeCallback(dsuid_t *id, int flag)
{
    try
    {
        FILE_LOG(logDEBUG1) << "dS485 Bus Change: " << util::str(*id) <<
                               " - " << flag;

        boost::shared_ptr<CheckBusMemberTypeTask> task(
                new CheckBusMemberTypeTask(*id, flag));
        addTask(boost::dynamic_pointer_cast<dSTask>(task));
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << "Exception in busChangeCallback: " <<
                              ex.getMessage();
    }
    catch (std::exception &e)
    {
        FILE_LOG(logERROR) << "Exception in busChangeCallback: " << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR) << "Exception in busChangeCallback";
    }
}

///****************************************************************************

void dS_Assistant::deviceAddedCallback(uint8_t error_code,
                                       dsuid_t src_dsm_dsuid,
                                       dsuid_t dest_dsm_dsuid,
                                       uint16_t device_id,
                                       uint16_t zone_id,
                                       uint16_t vendor_id,
                                       dsuid_t dsuid)
{
    try
    {
        boost::shared_ptr<Device> dsd =
                        Storage::getInstance()->getDevice(dsuid, src_dsm_dsuid);

        dsd->setState(1);
        Storage::getInstance()->addDevice(dsd);
    }
    catch (SQLite3Exception &sx)
    {
        FILE_LOG(logERROR) << "Could not load device " << util::str(dsuid) <<
            ": " << sx.getMessage();
        if (SQLite3::isFatal(sx.getErrorCode()))
        {
            FILE_LOG(logERROR) << "Unrecoverable SQLite error, terminating dSA";
            g_shutdown_flag = SHUTDOWN_SQLITE;
            return;
        }
    }
    catch (Exception &ex)
    {
        // exception is OK if device was not found
    }

    FILE_LOG(logINFO) << "Added device with dsuid " <<
            util::str(dsuid) << " and address " << util::str(device_id) <<
            " on dSM " << util::str(src_dsm_dsuid);

    boost::shared_ptr<GetDeviceInfoTask> task(
            new GetDeviceInfoTask(src_dsm_dsuid, device_id));
    addTask(boost::dynamic_pointer_cast<dSTask>(task));
}

void dS_Assistant::deviceRemovedCallback(uint8_t error_code,
                                         dsuid_t src_dsm_dsuid,
                                         dsuid_t dest_dsm_dsuid,
                                         uint16_t device_id,
                                         uint16_t zone_id,
                                         uint16_t vendor_id,
                                         dsuid_t dsuid)
{
    boost::shared_ptr<Device> dsd;
    try
    {
        dsd = Storage::getInstance()->getDevice(dsuid, src_dsm_dsuid);
        dsd->setState(0);
        Storage::getInstance()->addDevice(dsd);
    }
    catch (SQLite3Exception &sx)
    {
        FILE_LOG(logERROR) << "Could not load device " << util::str(dsuid) <<
            ": " << sx.getMessage();
        if (SQLite3::isFatal(sx.getErrorCode()))
        {
            FILE_LOG(logERROR) << "Unrecoverable SQLite error, terminating dSA";
            g_shutdown_flag = SHUTDOWN_SQLITE;
            return;
        }
    }
    catch (Exception &ex)
    {
        // exception is expected if device was not found
    }
    FILE_LOG(logINFO) << "Removed device with dsuid " <<
            util::str(dsuid) << " and address " << util::str(device_id) <<
            " on dSM " << util::str(src_dsm_dsuid);

    if (dsd != NULL)
    {
        DSUID_LOG(logINFO, util::str(dsd->getDSUID())) << "inactive";
    }
}

void dS_Assistant::deviceStatisticsCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
        uint8_t offset, uint16_t value)
{
    try
    {
        boost::shared_ptr<const Device> dsdo =
                Storage::getInstance()->getDevice(device_id, src_dsm_dsuid);
        DSUID_PREFIX_LOG(logDEBUG, dsdo->getDSUID(), "stat_") <<
            "statistic offset:" << util::hex(offset,2) <<
            ", value:" << util::hex(value,4);
    }
    catch (SQLite3Exception &sx)
    {
        FILE_LOG(logERROR) << "Could not load device : " << sx.getMessage();
        if (SQLite3::isFatal(sx.getErrorCode()))
        {
            FILE_LOG(logERROR) << "Unrecoverable SQLite error, terminating dSA";
            g_shutdown_flag = SHUTDOWN_SQLITE;
        }
    }
    catch (Exception &ex)
    {
        FILE_LOG(logWARNING) << "Device statistics callback: "
                             << ex.getMessage();
    }
}

void dS_Assistant::devicePingCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
        uint16_t device_id, uint8_t present)
{
}

void dS_Assistant::deviceActionCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
        uint16_t zone_id, uint8_t group_id, uint8_t button_nr,
        uint8_t click_type)
{
}

void dS_Assistant::deviceLocalActionCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
        uint16_t zone_id, uint8_t state)
{
}

void dS_Assistant::deviceConfigCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
        uint8_t config_class, uint8_t config_index, uint8_t value)
{
    try
    {
        boost::shared_ptr<const Device> dsdo =
                    Storage::getInstance()->getDevice(device_id, src_dsm_dsuid);
        DSUID_LOG(logINFO, util::str(dsdo->getDSUID())) <<
            "device config class:" << util::hex(config_class) <<
            ", index:" << util::hex(config_index, 2) <<
            ", value:" << util::hex(value, 4);
    }
    catch (SQLite3Exception &sx)
    {
        FILE_LOG(logERROR) << "Could not load device: " << sx.getMessage();
        if (SQLite3::isFatal(sx.getErrorCode()))
        {
            FILE_LOG(logERROR) << "Unrecoverable SQLite error, terminating dSA";
            g_shutdown_flag = SHUTDOWN_SQLITE;
        }
    }
    catch (Exception &ex)
    {
        FILE_LOG(logWARNING) << "Device config callback: "
                             << ex.getMessage();
    }
}

void dS_Assistant::deviceTestTransmissionQualityCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
        uint8_t downstream_quality, uint16_t upstream_quality)
{
    try
    {
        boost::shared_ptr<const Device> dsdo =
            Storage::getInstance()->getDevice(device_id, src_dsm_dsuid);
        DSUID_LOG(logINFO, util::str(dsdo->getDSUID())) <<
            "test quality down:" << util::hex(downstream_quality,2) <<
            ", up:" << util::hex(upstream_quality,2);
    }
    catch (SQLite3Exception &sx)
    {
        FILE_LOG(logERROR) << "Could not load device: " << sx.getMessage();
        if (SQLite3::isFatal(sx.getErrorCode()))
        {
            FILE_LOG(logERROR) << "Unrecoverable SQLite error, terminating dSA";
            g_shutdown_flag = SHUTDOWN_SQLITE;
        }
    }
    catch (Exception &ex)
    {
        FILE_LOG(logWARNING) << "Device test transmission quality callback: "
                             << ex.getMessage();
    }
}

///****************************************************************************

void dS_Assistant::commUpstreamShortCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
        uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
        uint8_t key_type, uint8_t click_type,
        uint8_t quality, uint8_t flags, uint8_t cflags)
{
}

void dS_Assistant::commUpstreamLongCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
        uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier,
        uint32_t data, uint8_t quality, uint8_t flags, uint8_t cflags)
{
}

void dS_Assistant::commDownstreamShortCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command)
{
}

void dS_Assistant::commDownstreamLong0Callback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command)
{
}

void dS_Assistant::commDownstreamLong8Callback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command,
        uint8_t data)
{
}

void dS_Assistant::commDownstreamLong16Callback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command,
        uint16_t data)
{
}

void dS_Assistant::dsmStatusCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
        uint8_t hardware, uint8_t communicationdsp, uint8_t temperature,
        uint8_t configuration, uint8_t ds485bus, uint8_t powerline)
{
}

void dS_Assistant::dsmErrorCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
        uint16_t error, uint32_t data)
{
}

void dS_Assistant::dsmRebootRequestCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid)
{
}

///****************************************************************************

void dS_Assistant::lockTaskList()
{
    int ret = pthread_mutex_lock(&m_task_list_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to lock task list mutex: " +
                         std::string(strerror(errno)));
    }
}

void dS_Assistant::unlockTaskList()
{
    int ret = pthread_mutex_unlock(&m_task_list_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to unlock task list mutex: " +
                         std::string(strerror(errno)));
    }
}

void dS_Assistant::lockLongTaskList() const
{
    int ret = pthread_mutex_lock(&m_long_task_list_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to lock long task list mutex: " +
                         std::string(strerror(errno)));
    }
}

void dS_Assistant::unlockLongTaskList() const
{
    int ret = pthread_mutex_unlock(&m_long_task_list_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to unlock long task list mutex: " +
                         std::string(strerror(errno)));
    }
}

void dS_Assistant::lockSuspendedTaskInfo() const
{
    int ret = pthread_mutex_lock(&m_suspended_tasks_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to lock suspended tasks mutex: " +
                         std::string(strerror(errno)));
    }
}

void dS_Assistant::unlockSuspendedTaskInfo() const
{
    int ret = pthread_mutex_unlock(&m_suspended_tasks_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to unlock suspended tasks mutex: " +
                         std::string(strerror(errno)));
    }
}


void dS_Assistant::run()
{
    int ret;

    if (m_running)
    {
        throw _Exception("Task threads are already running");
    }

    loadDatabaseDefaults();
    connect();

    m_running = true;

    try
    {
        clearPendingTasks();
    }
    catch (Exception &ex)
    {
        m_running = false;
        throw ex;
    }

    ret = pthread_create(&m_task_thread, NULL, dS_Assistant::staticThreadProc,
                         this);
    if (ret != 0)
    {
        disconnect();
        m_running = false;
        throw _Exception("Could not start task thread");
    }

    ret = pthread_create(&m_long_task_thread, NULL,
                         dS_Assistant::staticLongTaskThreadProc, this);
    if (ret != 0)
    {
        pthread_join(m_task_thread, NULL);
        disconnect();
        m_running = false;
        throw _Exception("Could not start long task thread");
    }

    try
    {
        m_comm_monitor_task = boost::shared_ptr<BasicMonitorTask>
            (new BasicMonitorTask());
        addTask(boost::dynamic_pointer_cast<dSTask>
            (m_comm_monitor_task));

        m_dsm_monitor_task = boost::shared_ptr<DsmMonitorTask>
            (new DsmMonitorTask());
        addTask(boost::dynamic_pointer_cast<dSTask>
            (m_dsm_monitor_task));

        int max_db_size;
        try
        {
            ConfigManager::getInstance()->getParameter(
                    "database.maxfilesize", max_db_size);
        }
        catch (Exception& ex) {
            max_db_size = DATABASE_MAX_FILESIZE;
        }

        m_db_cleanup_task = boost::shared_ptr<DatabaseCleanupTask>
            (new DatabaseCleanupTask(max_db_size));
        scheduleTask(boost::dynamic_pointer_cast<dSTask>
            (m_db_cleanup_task), 300);
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }

    try
    {
        Database::getInstance()->updateTaskStatus(Running, Cancelled);
        Database::getInstance()->updateDSMPresence(false);
        Database::getInstance()->activateDeviceDeleteFlag();

        boost::shared_ptr<std::list<boost::shared_ptr<dSLongTask> > > tasks =
            Database::getInstance()->getPersistentTasks();

        if (tasks->empty())
        {
            Database::getInstance()->removeTasksByStatus(Queued);
        }
        else
        {
            Database::getInstance()->removeTasksByStatus(Queued, false);
        }
        // no async startup anymore, because we have to queue the device
        // update task
        scan(false); // synchronous scan
        // remove all dSMs that are not present from the database
        Database::getInstance()->deleteDSMsByPresence(false);
        Database::getInstance()->deleteDevicesByDeleteFlag(true);

        // no persistent tasks
        if (tasks->empty())
        {
            dsuid_t devid;
            dsuid_t dsmid;
            SetNullDsuid(devid);
            SetNullDsuid(dsmid);
            boost::shared_ptr<DeviceUpdateTask> task(new DeviceUpdateTask(dsmid, devid));
            task->setPriority(Normal);
            addLongTask(task);
        }
        else
        {
            lockLongTaskList();
            while (!tasks->empty())
            {
                boost::shared_ptr<dSLongTask> task = tasks->front();
                tasks->pop_front();
                m_long_task_list.push_front(task);
            }
            pthread_cond_signal(&m_wakeup_long_task_cond);
            unlockLongTaskList();
        }
    }
    catch (SQLite3Exception &sx)
    {
        if (SQLite3::isFatal(sx.getErrorCode())) {
            throw sx;
        }
        FILE_LOG(logERROR) << sx.getMessage();
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
}

void dS_Assistant::taskThread()
{
    boost::shared_ptr<dSTask> task;
    bool terminate = false;

    lockTaskList();
    while (!m_shutdown_flag && !terminate)
    {
        if (m_task_list.empty())
        {
            pthread_cond_wait(&m_wakeup_task_cond, &m_task_list_mutex);
            continue;
        }

        task = m_task_list.back();
        m_task_list.pop_back();

        unlockTaskList();

        try
        {
            if ((task != NULL) && (task->isValid()))
            {
                FILE_LOG(logDEBUG1) << "running task " << task->getName();
                task->run(this);
            }
        }
        catch (SQLite3Exception &sx)
        {
            FILE_LOG(logERROR) << "Task " << task->getName()
                << " database exception: " <<
                sx.getMessage() << "[" << util::str(sx.getErrorCode()) << "]";
            sx.printStackTrace();
            if (SQLite3::isFatal(sx.getErrorCode()))
            {
                FILE_LOG(logERROR) << "Unrecoverable SQLite error, terminating";
                terminate = true;
            }
        }
        catch (Exception &ex)
        {
            FILE_LOG(logERROR) << "Task " << task->getName()
                << " exception: " << ex.getMessage();
            ex.printStackTrace();
        }
        catch (std::exception &e)
        {
            FILE_LOG(logERROR) << "Task " << task->getName()
                << " std::exception: " << e.what();
        }
        catch (...)
        {
            FILE_LOG(logERROR) << "Task " << task->getName()
                << " unknown exception";
        }

        task.reset();

        if (!m_shutdown_flag && !terminate)
        {
            lockTaskList();
        }
    }
    unlockTaskList();
    if (terminate)
    {
        FILE_LOG(logERROR) << "Unrecoverable SQLite error, terminating dSA";
        g_shutdown_flag = SHUTDOWN_SQLITE;
    }
}

void *dS_Assistant::staticThreadProc(void *arg)
{
    dS_Assistant *instance = (dS_Assistant *)arg;
    instance->taskThread();
    pthread_exit(NULL);
    return NULL;
}

void dS_Assistant::longTaskThread()
{
    boost::shared_ptr<dSLongTask> task;
    bool terminate = false;

    lockLongTaskList();
    while (!m_shutdown_flag && !terminate)
    {
        if (m_paused_task_list.empty() &&
            m_long_task_list.empty() &&
            m_high_prio_task_list.empty())
        {
            FILE_LOG(logDEBUG3) << "no long tasks pending, sleeping...";
            pthread_cond_wait(&m_wakeup_long_task_cond,
                              &m_long_task_list_mutex);
            continue;
        }

        // high prio tasks will always be sorted in front by the add function
        if (!m_high_prio_task_list.empty())
        {
            FILE_LOG(logDEBUG3) << "queueing high priority task";
            task = m_high_prio_task_list.back();
            m_high_prio_task_list.pop_back();
        }

        // if no high priority task is queued, proceed with the paused tasks
        if ((task == NULL) && (!m_paused_task_list.empty()))
        {
            FILE_LOG(logDEBUG3) << "queueing paused task";
            task = m_paused_task_list.back();
            m_paused_task_list.pop_back();
        }

        if (task == NULL)
        {
            while (!m_long_task_list.empty())
            {
                boost::shared_ptr<dSLongTask> temp = m_long_task_list.back();
                m_long_task_list.pop_back();
                if (!temp->isPersistent())
                {
                    task = temp;
                    FILE_LOG(logDEBUG3) << "queueing regular task";
                    break;
                }
                else
                {
                    if (!isTaskSuspended(temp->getTaskType()))
                    {
                        task = temp;
                        FILE_LOG(logDEBUG3) << "queueing persistent task";
                        break;
                    }
                }
            }
        }

        // long task list contained only suspended tasks
        if (task == NULL)
        {
            continue;
        }

        m_current_task_info = boost::dynamic_pointer_cast<dSLongTaskInfo>(task);
        unlockLongTaskList();

        try
        {
            if (task->isValid())
            {
                FILE_LOG(logDEBUG1) << "running long task " << task->getName();
                Database::getInstance()->updateTaskStatus(task->getTaskID(),
                                                          Running);
                task->run(this);
                Database::getInstance()->updateTaskStatus(task->getTaskID(),
                                                        task->getTaskStatus());

                if (task->isValid() && task->isPaused() &&
                        (!task->isPersistent() ||
                            (task->isPersistent() &&
                            !isTaskSuspended(task->getTaskType()))))
                {
                    m_paused_task_list.push_front(task);
                }
            }
        }
        catch (SQLite3Exception &sx)
        {
            FILE_LOG(logERROR) << "Long task " << task->getName()
                << " database exception: " << sx.getMessage();
            sx.printStackTrace();
            if (SQLite3::isFatal(sx.getErrorCode()))
            {
                terminate = true;
            }
        }
        catch (Exception &ex)
        {
            FILE_LOG(logERROR) << "Long task " << task->getName()
                << " exception: " << ex.getMessage();
            ex.printStackTrace();
        }
        catch (std::exception &e)
        {
            FILE_LOG(logERROR) << "Long task " << task->getName()
                << " std::exception: " << e.what();
        }
        catch (...)
        {
            FILE_LOG(logERROR) << "Long task " << task->getName()
                << " unknown exception";
        }

        task.reset();

        if (!m_shutdown_flag && !terminate)
        {
            lockLongTaskList();
            m_current_task_info.reset();
        }
    }
    unlockLongTaskList();

    if (terminate)
    {
        FILE_LOG(logERROR) << "Unrecoverable SQLite error, terminating dSA";
        g_shutdown_flag = SHUTDOWN_SQLITE;
    }
}

void *dS_Assistant::staticLongTaskThreadProc(void *arg)
{
    dS_Assistant *instance = (dS_Assistant *)arg;
    instance->longTaskThread();
    pthread_exit(NULL);
    return NULL;
}

void dS_Assistant::scan(bool async)
{
    Database::getInstance()->updateDSMPresence(false);
    boost::shared_ptr<GetBusMembersTask> task(new GetBusMembersTask(async));
    if (async)
    {
        addTask(boost::dynamic_pointer_cast<dSTask>(task));
    }
    else
    {
        task->run(this);
    }

    boost::shared_ptr<CheckDSMThread> dsmtask(new CheckDSMThread(1800));
    scheduleLongTask(boost::dynamic_pointer_cast<dSLongTask>(dsmtask), 60);
}

void dS_Assistant::clearPendingTasks()
{
    lockTaskList();
    while (!m_task_list.empty())
    {
        m_task_list.pop_front();
    }
    unlockTaskList();

    lockLongTaskList();
    while (!m_high_prio_task_list.empty())
    {
        m_high_prio_task_list.pop_front();
    }

    while (!m_long_task_list.empty())
    {
        m_long_task_list.pop_front();
    }

    while (!m_paused_task_list.empty())
    {
        m_paused_task_list.pop_front();
    }

    try
    {
        while (!m_scheduled_task_list.empty())
        {
            boost::weak_ptr<LongTaskScheduler> tw = m_scheduled_task_list.front();
            if (boost::shared_ptr<LongTaskScheduler> t = tw.lock())
            {
                Timer::getInstance()->removeSubscriber(t);
            }
            m_scheduled_task_list.pop_front();
        }
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    unlockLongTaskList();
}

boost::shared_ptr<const dSLongTaskInfo> dS_Assistant::getCurrentTaskInfo() const
{
    lockLongTaskList();
    boost::shared_ptr<const dSLongTaskInfo> ret = m_current_task_info;
    unlockLongTaskList();
    return ret;
}

boost::shared_ptr<std::list<boost::shared_ptr<const dSLongTaskInfo> > >
                                        dS_Assistant::getTaskList() const
{
    lockLongTaskList();
    boost::shared_ptr<std::list<boost::shared_ptr<const dSLongTaskInfo> > >
        taskList(new std::list<boost::shared_ptr<const dSLongTaskInfo> >);

    if (m_current_task_info != NULL)
    {
        taskList->push_back(m_current_task_info);
    }

    for (std::list<boost::shared_ptr<dSLongTask> >::const_iterator
            it = m_high_prio_task_list.begin();
            it != m_high_prio_task_list.end(); it++)
    {
        boost::shared_ptr<const dSLongTaskInfo> task =
                            boost::dynamic_pointer_cast<dSLongTaskInfo>(*it);
        taskList->push_back(task);
    }

    for (std::list<boost::shared_ptr<dSLongTask> >::const_iterator
            it = m_paused_task_list.begin();
            it != m_paused_task_list.end(); it++)
    {
        boost::shared_ptr<const dSLongTaskInfo> task =
                            boost::dynamic_pointer_cast<dSLongTaskInfo>(*it);
        taskList->push_back(task);
    }

    for (std::list<boost::shared_ptr<dSLongTask> >::const_iterator
            it = m_long_task_list.begin(); it != m_long_task_list.end(); it++)
    {
        boost::shared_ptr<const dSLongTaskInfo> task =
                            boost::dynamic_pointer_cast<dSLongTaskInfo>(*it);
        taskList->push_back(task);
    }

    for (std::list<boost::shared_ptr<dSLongTask> >::const_iterator
            it = m_suspended_task_list.begin(); it != m_suspended_task_list.end(); it++)
    {
        boost::shared_ptr<const dSLongTaskInfo> task =
                            boost::dynamic_pointer_cast<dSLongTaskInfo>(*it);
        taskList->push_back(task);
    }

    for (std::list<boost::weak_ptr<LongTaskScheduler> >::const_iterator
            it = m_scheduled_task_list.begin();
            it != m_scheduled_task_list.end(); it++)
    {
        boost::shared_ptr<LongTaskScheduler> s = (*it).lock();
        if (s)
        {
            boost::shared_ptr<const dSLongTaskInfo> task = s->getTask();
            taskList->push_back(task);
        }
    }

    unlockLongTaskList();

    return taskList;
}

boost::shared_ptr<dSLongTaskInfo> dS_Assistant::getTaskInfo(long long int task_id)
{
    boost::shared_ptr<dSLongTaskInfo> task;

    lockLongTaskList();

    if (m_current_task_info && (m_current_task_info->getTaskID() == task_id))
    {
        task = m_current_task_info;
        unlockLongTaskList();
        return task;
    }

    for (std::list<boost::shared_ptr<dSLongTask> >::const_iterator
            it = m_high_prio_task_list.begin();
            it != m_high_prio_task_list.end(); it++)
    {
        if ((*it)->getTaskID() == task_id)
        {
            task = (*it);
            unlockLongTaskList();
            return task;
        }
    }

    for (std::list<boost::shared_ptr<dSLongTask> >::const_iterator
            it = m_paused_task_list.begin();
            it != m_paused_task_list.end(); it++)
    {
        if ((*it)->getTaskID() == task_id)
        {
            task = (*it);
            unlockLongTaskList();
            return task;
        }
    }

    for (std::list<boost::shared_ptr<dSLongTask> >::const_iterator
            it = m_suspended_task_list.begin();
            it != m_suspended_task_list.end(); it++)
    {
        if ((*it)->getTaskID() == task_id)
        {
            task = (*it);
            unlockLongTaskList();
            return task;
        }
    }

    for (std::list<boost::shared_ptr<dSLongTask> >::const_iterator
            it = m_long_task_list.begin();
            it != m_long_task_list.end(); it++)
    {
        if ((*it)->getTaskID() == task_id)
        {
            task = (*it);
            unlockLongTaskList();
            return task;
        }
    }

    for (std::list<boost::weak_ptr<LongTaskScheduler> >::const_iterator
            it = m_scheduled_task_list.begin();
            it != m_scheduled_task_list.end(); it++)
    {
        boost::shared_ptr<LongTaskScheduler> s = (*it).lock();
        if (s)
        {
            boost::shared_ptr<dSLongTaskInfo> t = s->getTask();
            if (t->getTaskID() == task_id)
            {
                task = t;
                unlockLongTaskList();
                return task;
            }
        }
    }

    unlockLongTaskList();
    return boost::shared_ptr<dSLongTaskInfo> ();
}

boost::shared_ptr<dSLongTaskInfo> dS_Assistant::getTaskInfoByType(task_type task_type)
{
    boost::shared_ptr<dSLongTaskInfo> task;

    lockLongTaskList();

    if (m_current_task_info && (m_current_task_info->getTaskType() == task_type))
    {
        task = m_current_task_info;
        unlockLongTaskList();
        return task;
    }

    for (std::list<boost::shared_ptr<dSLongTask> >::const_iterator
            it = m_high_prio_task_list.begin();
            it != m_high_prio_task_list.end(); it++)
    {
        if ((*it)->getTaskType() == task_type)
        {
            task = (*it);
            unlockLongTaskList();
            return task;
        }
    }

    for (std::list<boost::shared_ptr<dSLongTask> >::const_iterator
            it = m_paused_task_list.begin();
            it != m_paused_task_list.end(); it++)
    {
        if ((*it)->getTaskType() == task_type)
        {
            task = (*it);
            unlockLongTaskList();
            return task;
        }
    }

    for (std::list<boost::shared_ptr<dSLongTask> >::const_iterator
            it = m_suspended_task_list.begin();
            it != m_suspended_task_list.end(); it++)
    {
        if ((*it)->getTaskType() == task_type)
        {
            task = (*it);
            unlockLongTaskList();
            return task;
        }
    }

    for (std::list<boost::shared_ptr<dSLongTask> >::const_iterator
            it = m_long_task_list.begin();
            it != m_long_task_list.end(); it++)
    {
        if ((*it)->getTaskType() == task_type)
        {
            task = (*it);
            unlockLongTaskList();
            return task;
        }
    }

    for (std::list<boost::weak_ptr<LongTaskScheduler> >::const_iterator
            it = m_scheduled_task_list.begin();
            it != m_scheduled_task_list.end(); it++)
    {
        boost::shared_ptr<LongTaskScheduler> s = (*it).lock();
        if (s)
        {
            boost::shared_ptr<dSLongTaskInfo> t = s->getTask();
            if (t->getTaskType() == task_type)
            {
                task = t;
                unlockLongTaskList();
                return task;
            }
        }
    }

    unlockLongTaskList();
    return boost::shared_ptr<dSLongTaskInfo> ();
}

boost::shared_ptr<dSLongTaskInfo> dS_Assistant::getTaskInfoByName(const std::string& name)
{
    boost::shared_ptr<dSLongTaskInfo> task;

    lockLongTaskList();

    if (m_current_task_info && (m_current_task_info->getName() == name))
    {
        task = m_current_task_info;
        unlockLongTaskList();
        return task;
    }

    for (std::list<boost::shared_ptr<dSLongTask> >::const_iterator
            it = m_high_prio_task_list.begin();
            it != m_high_prio_task_list.end(); it++)
    {
        if ((*it)->getName() == name)
        {
            task = (*it);
            unlockLongTaskList();
            return task;
        }
    }

    for (std::list<boost::shared_ptr<dSLongTask> >::const_iterator
            it = m_paused_task_list.begin();
            it != m_paused_task_list.end(); it++)
    {
        if ((*it)->getName() == name)
        {
            task = (*it);
            unlockLongTaskList();
            return task;
        }
    }

    for (std::list<boost::shared_ptr<dSLongTask> >::const_iterator
            it = m_suspended_task_list.begin();
            it != m_suspended_task_list.end(); it++)
    {
        if ((*it)->getName() == name)
        {
            task = (*it);
            unlockLongTaskList();
            return task;
        }
    }

    for (std::list<boost::shared_ptr<dSLongTask> >::const_iterator
            it = m_long_task_list.begin();
            it != m_long_task_list.end(); it++)
    {
        if ((*it)->getName() == name)
        {
            task = (*it);
            unlockLongTaskList();
            return task;
        }
    }

    for (std::list<boost::weak_ptr<LongTaskScheduler> >::const_iterator
            it = m_scheduled_task_list.begin();
            it != m_scheduled_task_list.end(); it++)
    {
        boost::shared_ptr<LongTaskScheduler> s = (*it).lock();
        if (s)
        {
            boost::shared_ptr<dSLongTaskInfo> t = s->getTask();
            if (t->getName() == name)
            {
                task = t;
                unlockLongTaskList();
                return task;
            }
        }
    }

    unlockLongTaskList();
    return boost::shared_ptr<dSLongTaskInfo> ();
}

void dS_Assistant::pauseTask(long long int task_id)
{
    bool found = false;
    lockLongTaskList();

    if ((m_current_task_info != NULL) &&
        ((m_current_task_info->getTaskID() == task_id) ||
         (m_current_task_info->getParentTaskID() == task_id)))
    {
        found = true;
        if (m_current_task_info->isPausable())
        {
            m_current_task_info->pause();
        }
        else
        {
            unlockLongTaskList();
            throw _Exception(_("task does not support pausing"));
        }
    }

    try
    {
        if (found)
        {
            Database::getInstance()->updateTaskStatus(task_id, Cancelled);
        }
    }
    catch (Exception &ex)
    {
        FILE_LOG(logDEBUG) << ex.getMessage();
        unlockLongTaskList();
        throw ex;
    }

    unlockLongTaskList();

    if (!found)
    {
        throw _Exception(_P(_("task with id %lld not found"),
                                     task_id));
    }
}

void dS_Assistant::cancelTask(long long int task_id)
{
    bool found = false;

    std::list<long long int> parents;

    lockLongTaskList();
    try
    {
        std::list<boost::shared_ptr<dSLongTask> >::iterator it =
            m_long_task_list.begin();

        while (it != m_long_task_list.end())
        {
            boost::shared_ptr<dSLongTask> task = *it;
            if ((task->getTaskID() == task_id) ||
                (task->getParentTaskID() == task_id))
            {
                found = true;
                task->invalidate();
                it = m_long_task_list.erase(it);
                if (task->getParentTaskID() != INVALID_TASK_ID)
                {
                    parents.push_front(task->getParentTaskID());
                }
                continue;
            }
            it++;
        }

        it = m_high_prio_task_list.begin();
        while (it != m_high_prio_task_list.end())
        {
            boost::shared_ptr<dSLongTask> task = *it;
            if ((task->getTaskID() == task_id) ||
                (task->getParentTaskID() == task_id))
            {
                found = true;
                task->invalidate();
                it = m_high_prio_task_list.erase(it);
                if (task->getParentTaskID() != INVALID_TASK_ID)
                {
                    parents.push_front(task->getParentTaskID());
                }
                continue;
            }
            it++;
        }

        it = m_paused_task_list.begin();
        while (it != m_paused_task_list.end())
        {
            boost::shared_ptr<dSLongTask> task = *it;
            if ((task->getTaskID() == task_id) ||
                (task->getParentTaskID() == task_id))
            {
                found = true;
                task->invalidate();
                it = m_paused_task_list.erase(it);
                if (task->getParentTaskID() != INVALID_TASK_ID)
                {
                    parents.push_front(task->getParentTaskID());
                }
                continue;
            }
            it++;
        }

        std::list<boost::weak_ptr<LongTaskScheduler> >::iterator wi =
            m_scheduled_task_list.begin();
        while (wi != m_scheduled_task_list.end())
        {
            boost::weak_ptr<LongTaskScheduler> tw = *wi;
            if (boost::shared_ptr<LongTaskScheduler> ts = tw.lock())
            {
                if ((ts->getTask()->getTaskID() == task_id) ||
                    (ts->getTask()->getParentTaskID() == task_id))
                {
                    found = true;
                    ts->getTask()->invalidate();
                    Timer::getInstance()->removeSubscriber(ts);
                    wi = m_scheduled_task_list.erase(wi);
                    if (ts->getTask()->getParentTaskID() != INVALID_TASK_ID)
                    {
                        parents.push_front(ts->getTask()->getParentTaskID());
                    }
                    continue;
                }
            }
            else
            {
                wi = m_scheduled_task_list.erase(wi);
                continue;
            }
            wi++;
        }

        if ((m_current_task_info != NULL) &&
            ((m_current_task_info->getTaskID() == task_id) ||
             (m_current_task_info->getParentTaskID() == task_id)))
        {
            found = true;
            if (m_current_task_info->getParentTaskID() != INVALID_TASK_ID)
            {
                parents.push_front(m_current_task_info->getParentTaskID());
            }
            m_current_task_info->invalidate();
        }

        if (found)
        {
            Database::getInstance()->updateTaskStatus(task_id, Cancelled);
        }
    }
    catch (Exception &ex)
    {
        FILE_LOG(logDEBUG) << ex.getMessage();
        unlockLongTaskList();
        throw ex;
    }
    catch (std::runtime_error &err)
    {
        FILE_LOG(logERROR) << err.what();
    }
    unlockLongTaskList();

    if (!found)
    {
        throw _Exception("task with id " + util::str(task_id) + " not found");
    }

    parents.unique();
    while (!parents.empty())
    {
        try
        {
            this->cancelTask(parents.front());
        }
        catch (Exception &ex)
        {
            // parent tasks may not be around anymore
            FILE_LOG(logDEBUG2) << " OK if task now found: " << ex.getMessage();
        }
        parents.pop_front();
    }
}

void dS_Assistant::suspendTask(task_type type)
{
    long long int task_id = INVALID_TASK_ID;
    lockLongTaskList();
    try
    {
        if (m_current_task_info != NULL)
        {
            if (m_current_task_info->getTaskType() == type)
            {
                if (!m_current_task_info->isPersistent())
                {
                    throw _Exception("only persistent tasks can be suspended");
                }

                task_id = m_current_task_info->getTaskID();
                if (m_current_task_info->isPausable())
                {
                    if (!isTaskSuspended(m_current_task_info->getTaskType()))
                    {
                        setSuspendedTask(m_current_task_info->getTaskType());
                        m_suspended_task_list.push_back(boost::dynamic_pointer_cast<dSLongTask>(m_current_task_info));
                        boost::dynamic_pointer_cast<dSLongTask>(m_current_task_info)->setTaskStatus(Suspended);
                    }
                    m_current_task_info->pause();
                    m_current_task_info.reset();
                }
                else
                {
                    throw _Exception(_("task can not be paused / suspended"));
                }
            }
        }

        std::list<boost::shared_ptr<dSLongTask> >::iterator it =
            m_long_task_list.begin();

        while ((task_id == INVALID_TASK_ID) && it != m_long_task_list.end())
        {
            boost::shared_ptr<dSLongTask> task = *it;
            if (task->getTaskType() == type)
            {
                if (!task->isPersistent())
                {
                    throw _Exception("only persistent tasks can be suspended");
                }

                task_id = task->getTaskID();
                if (task->isPausable())
                {
                    it = m_long_task_list.erase(it);

                    if (!isTaskSuspended(task->getTaskType()))
                    {
                        setSuspendedTask(task->getTaskType());
                        m_suspended_task_list.push_back(task);
                    }
                    task->pause();
                    task->setTaskStatus(Suspended);
                }
                else
                {
                    throw _Exception(_("task can not be paused / suspended"));
                }

                break;
            }
            it++;
        }

        it = m_paused_task_list.begin();
        while ((task_id == INVALID_TASK_ID) && it != m_paused_task_list.end())
        {
            boost::shared_ptr<dSLongTask> task = *it;
            if (task->getTaskType() == type)
            {
                if (!task->isPersistent())
                {
                    throw _Exception("only persistent tasks can be suspended");
                }

                task_id = task->getTaskID();
                if (task->isPausable())
                {
                    it = m_paused_task_list.erase(it);
                    if (!isTaskSuspended(task->getTaskType()))
                    {
                        setSuspendedTask(task->getTaskType());
                        m_suspended_task_list.push_back(task);
                    }
                    task->pause();
                    task->setTaskStatus(Suspended);
                }
                else
                {
                    throw _Exception(_("task can not be paused / suspended"));
                }

                break;
            }
            it++;
        }

        std::list<boost::weak_ptr<LongTaskScheduler> >::iterator wi =
            m_scheduled_task_list.begin();
        while ((task_id == INVALID_TASK_ID) && wi != m_scheduled_task_list.end())
        {
            boost::weak_ptr<LongTaskScheduler> tw = *wi;
            if (boost::shared_ptr<LongTaskScheduler> ts = tw.lock())
            {
                if (ts->getTask()->getTaskType() == type)
                {
                    if (!ts->getTask()->isPersistent())
                    {
                        throw _Exception("only persistent tasks can be suspended");
                    }

                    task_id = ts->getTask()->getTaskID();

                    if (ts->getTask()->isPausable())
                    {
                        Timer::getInstance()->removeSubscriber(ts);
                        wi = m_scheduled_task_list.erase(wi);
                        if (!isTaskSuspended(ts->getTask()->getTaskType()))
                        {
                            setSuspendedTask(ts->getTask()->getTaskType());
                            m_suspended_task_list.push_back(ts->getTask());
                        }
                        ts->getTask()->pause();
                        ts->getTask()->setTaskStatus(Suspended);
                    }
                    else
                    {
                        throw _Exception(_("task can not be paused / suspended"));
                    }

                    break;
                }
            }
            wi++;
        }

        if (task_id != INVALID_TASK_ID)
        {
            Database::getInstance()->updateTaskStatus(task_id, Suspended);
        }
    }
    catch (Exception &ex)
    {
        FILE_LOG(logDEBUG) << ex.getMessage();
        unlockLongTaskList();
        throw ex;
    }
    catch (std::runtime_error &err)
    {
        FILE_LOG(logERROR) << err.what();
    }
    unlockLongTaskList();

    if (!task_id)
    {
        throw _Exception(_P(_("task to suspend with type %d not found"),
                               type));
    }
}

void dS_Assistant::resumeTask(task_type type)
{
    bool found = false;

    lockLongTaskList();
    try
    {
        std::list<boost::shared_ptr<dSLongTask> >::iterator it =
            m_suspended_task_list.begin();

        while (it != m_suspended_task_list.end())
        {
            boost::shared_ptr<dSLongTask> task = *it;
            if (task->getTaskType() == type)
            {
                found = true;
                it = m_suspended_task_list.erase(it);
                unsetSuspendedTask(task->getTaskType());
                task->setTaskStatus(Queued);
                if (task->getQueuedTime() > 0)
                {
                    unlockLongTaskList();
                    scheduleLongTask(task, task->getQueuedTime());
                }
                else
                {
                    unlockLongTaskList();
                    addLongTask(task);
                }
                break;
            }
            it++;
        }
    }
    catch (Exception &ex)
    {
        FILE_LOG(logDEBUG) << ex.getMessage();
        unlockLongTaskList();
        throw ex;
    }
    catch (std::runtime_error &err)
    {
        FILE_LOG(logERROR) << err.what();
    }

    if (!found)
    {
        unlockLongTaskList();
        throw _Exception(_P(_("task to resume with type %d not found"), type));
    }
}

void dS_Assistant::disconnect()
{
    boost::mutex::scoped_lock lock(m_dsa_mutex);
    if (m_dsm_api_wrapper)
    {
        try
        {
            clearPendingTasks();
        }
        catch (Exception &ex)
        {
            FILE_LOG(logERROR) << "Failed to clear pending tasks: " <<
                ex.getMessage();
        }
        m_dsm_api_wrapper->removeListener(this);
        m_dsm_api_wrapper.reset();
    }
}

void dS_Assistant::shutdown()
{
    boost::mutex::scoped_lock lock(m_dsa_mutex);
    m_shutdown_flag = true;

    clearPendingTasks();
    lockTaskList();
    pthread_cond_signal(&m_wakeup_task_cond);
    unlockTaskList();
    if (m_task_thread)
    {
        pthread_join(m_task_thread, NULL);
        m_task_thread = 0;
    }

    lockLongTaskList();
    if (m_current_task_info != NULL)
    {
        m_current_task_info->invalidate();
    }
    pthread_cond_signal(&m_wakeup_long_task_cond);
    unlockLongTaskList();
    if (m_long_task_thread)
    {
        pthread_join(m_long_task_thread, NULL);
        m_long_task_thread = 0;
    }

    lock.unlock();
    disconnect();
}

const dsuid_t dS_Assistant::getDSUID() const
{
    return m_own_dsuid;
}

long long int dS_Assistant::getInvalidTaskID() const
{
    return INVALID_TASK_ID;
}
dS_Assistant::~dS_Assistant()
{
    try
    {
        shutdown();
        disconnect();
        Database::getInstance()->updateTaskStatus(Running, Cancelled);
        Database::getInstance()->updateDSMPresence(false);
        Database::getInstance()->removeTasksByStatus(Queued, false);
    }
    catch (...)
    {
        FILE_LOG(logERROR) << "Exception caught while shutting down, ignoring";
    }

    boost::mutex::scoped_lock lock(m_dsa_mutex);
    m_instance = NULL;
    lock.unlock();

    pthread_mutex_destroy(&m_task_list_mutex);
    pthread_mutex_destroy(&m_long_task_list_mutex);
    pthread_cond_destroy(&m_wakeup_task_cond);
    pthread_cond_destroy(&m_wakeup_long_task_cond);
    pthread_mutex_destroy(&m_suspended_tasks_mutex);
}



} // namespace
